<html>
  <head>
    <title>{{ visit[0]['Visit name'] }}</title>
    <link href="{{ url_for('static', path='/styles.css') }}" rel="stylesheet" />
    <link
      href="{{ url_for('images', path='/icon_268.png') }}"
      rel="icon"
      type="image/png"
    />
  </head>

  <body>
    <div class="topnav">
      <a href="/">Home</a>
      <a href="/visits">Active Visits</a>
    </div>

    <div class="content">
      <h1>{{ visit[0]['Visit name'] }}</h1>
    </div>

    <div class="visits">
      <p>Time remaining: {{ visit[0]['Time remaining'] }}</p>
      <p>Start time: {{ visit[0]['Start date'] }}</p>
      <p>End time: {{ visit[0]['End date'] }}</p>
    </div>

    <div class="footer">
      <p>For help please contact your LC or the EM Data Analysis team.</p>
    </div>
  </body>
</html>
