{% extends "base.html" %} {% block title %}MURFEY{% endblock %} {% block content
%}
<h1>{{ microscope }}</h1>
<p>
  A transporter for data from Diamond eBIC microscope and detector machines onto
  the Diamond network.
</p>
{% endblock %}
